 /**
 ********************************************************************************
 * Copyright (c) 2017 - 2018 Robert Bosch GmbH.
 * 
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Robert Bosch GmbH - initial API and implementation
 ********************************************************************************
 */
package org.eclipse.app4mc.emfutils.content.extractor.builders;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.eclipse.emf.common.util.BasicEMap;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EStructuralFeature.Setting;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.util.EcoreUtil;

public class EClassAllReferencesBufferBuilder {


	private EMap<String, Object> id2ObjectsMap;

	public EMap<String, Object> getId2ObjectsMap() {
		return this.id2ObjectsMap;
	}


	public void setId2ObjectsMap(final EMap<String, Object> id2ObjectsMap) {
		this.id2ObjectsMap = id2ObjectsMap;
	}


	public StringBuffer buildBuffer(final EClass selectedObjClass) {

		this.id2ObjectsMap = new BasicEMap<String, Object>();


		final StringBuffer plantumlBuffer = new StringBuffer();


		plantumlBuffer.append("@startuml");
		plantumlBuffer.append(System.getProperty("line.separator"));


		addClassInfo(selectedObjClass, plantumlBuffer);


		populateHeirarchy(plantumlBuffer, selectedObjClass);

		plantumlBuffer.append(System.getProperty("line.separator"));

		/*- also display information of sub-classes */
		populateSubClasses(plantumlBuffer, selectedObjClass);


		plantumlBuffer.append(System.getProperty("line.separator"));
		plantumlBuffer.append("@enduml");

		return plantumlBuffer;

	}


	private void addClassInfo(final EClass selectedObjClass, final StringBuffer plantumlBuffer) {
		plantumlBuffer.append(System.getProperty("line.separator"));


		if (selectedObjClass.isAbstract() && selectedObjClass.isInterface()) {
			plantumlBuffer.append("interface \"" + selectedObjClass.getName() + "\"");

		}
		else if (selectedObjClass.isAbstract()) {
			plantumlBuffer.append("abstract \"" + selectedObjClass.getName() + "\"");

		}
		else if (selectedObjClass.isInterface()) {

			plantumlBuffer.append("interface \"" + selectedObjClass.getName() + "\"");
		}
		else {
			plantumlBuffer.append("class \"" + selectedObjClass.getName() + "\"");

		}

		plantumlBuffer.append(System.getProperty("line.separator"));
	}


	private void populateHeirarchy(final StringBuffer plantumlBuffer, final EClass selectedObjClass) {

		populateEClassReferences(plantumlBuffer, selectedObjClass);

		for (final EClass superClass : selectedObjClass.getESuperTypes()) {


			if (superClass.isAbstract() && superClass.isInterface()) {
				addContentToBuffer(plantumlBuffer,
						"interface \"" + superClass.getName() + "\"" + "    [[" + getUUID(superClass) + "]]");

			}
			else if (superClass.isAbstract()) {
				addContentToBuffer(plantumlBuffer,
						"abstract \"" + superClass.getName() + "\"" + "    [[" + getUUID(superClass) + "]]");

			}
			else if (superClass.isInterface()) {

				addContentToBuffer(plantumlBuffer,
						"interface \"" + superClass.getName() + "\"" + "    [[" + getUUID(superClass) + "]]");
			}
			else {
				addContentToBuffer(plantumlBuffer,
						"class \"" + superClass.getName() + "\"" + "    [[" + getUUID(superClass) + "]]");

			}

			plantumlBuffer.append(System.getProperty("line.separator"));

			addContentToBuffer(plantumlBuffer,
					superClass.getName() + " <|-- " + selectedObjClass.getName() + "  #darkblue");

			plantumlBuffer.append(System.getProperty("line.separator"));

			populateHeirarchy(plantumlBuffer, superClass);


		}


	}

	private void addContentToBuffer(final StringBuffer plantumlBuffer, final String data) {

		if (plantumlBuffer.toString().contains(data) == false) {
			plantumlBuffer.append(data);
		}

	}

	private void populateSubClasses(final StringBuffer plantumlBuffer, final EClass selectedObjClass) {

		for (final EClass subClass : getSubClasses(selectedObjClass)) {


			if (subClass.isAbstract() && subClass.isInterface()) {


				addContentToBuffer(plantumlBuffer,
						"interface \"" + subClass.getName() + "\"" + "    [[" + getUUID(subClass) + "]]");

			}
			else if (subClass.isAbstract()) {
				addContentToBuffer(plantumlBuffer,
						"abstract \"" + subClass.getName() + "\"" + "    [[" + getUUID(subClass) + "]]");

			}
			else if (subClass.isInterface()) {

				addContentToBuffer(plantumlBuffer,
						"interface \"" + subClass.getName() + "\"" + "    [[" + getUUID(subClass) + "]]");
			}
			else {
				addContentToBuffer(plantumlBuffer,
						"class \"" + subClass.getName() + "\"" + "    [[" + getUUID(subClass) + "]]");

			}

			plantumlBuffer.append(System.getProperty("line.separator"));

			addContentToBuffer(plantumlBuffer,
					selectedObjClass.getName() + " <|-- " + subClass.getName() + "  #darkGreen");

			plantumlBuffer.append(System.getProperty("line.separator"));

			populateEClassContainerInfo(plantumlBuffer, subClass);

			populateSubClasses(plantumlBuffer, subClass);

		}
	}


	private List<EClass> getSubClasses(final EClass selectedObjClass) {

		final List<EClass> allSubClasses = new ArrayList<EClass>();

		final ResourceSet resourceSet = selectedObjClass.eResource().getResourceSet();

		if (resourceSet != null) {

			// With this approach check is made to find the Sub-Classes for the supplied EClass in all the packages
			// belonging to the resourceset

			final EList<Resource> resources = resourceSet.getResources();

			for (final Resource resource : resources) {

				final EList<EObject> contents = resource.getContents();

				if (contents.size() > 0) {
//					final EObject eObject = contents.get(0);

					for (EObject eobj : contents) {
						
						if (eobj instanceof EPackage) {
							allSubClasses.addAll(getSubClasses(selectedObjClass, (EPackage) eobj));
						}
					}
				}

			}
		}
		else {

			final EPackage ePackage = selectedObjClass.getEPackage();
			return getSubClasses(selectedObjClass, ePackage);
		}

		return allSubClasses;

	}

	private List<EClass> getSubClasses(final EClass selectedObjClass, final EPackage ePackage) {
		final List<EClass> allSubClasses = new ArrayList<EClass>();

		final EList<EClassifier> eClassifiers = ePackage.getEClassifiers();

		for (final EClassifier eClassifier : eClassifiers) {

			if (eClassifier instanceof EClass) {

				if (!(selectedObjClass == eClassifier) && selectedObjClass.isSuperTypeOf((EClass) eClassifier)) {
					allSubClasses.add((EClass) eClassifier);
				}


			}
		}
		return allSubClasses;
	}


	private void populateEClassReferences(final StringBuffer plantumlBuffer, final EClass selectedObjClass) {
		final List<EReference> result = new ArrayList<EReference>();

		final Object searchScope = (selectedObjClass.eResource().getResourceSet() == null)
				? selectedObjClass.eResource() : selectedObjClass.eResource().getResourceSet();

		/*- getting all the elements from the selected objects EClass --> by supplying the EResource of the corresponding EClass (i.e. ecore or Xcore) */

		Collection<Setting> allRefSettings = null;
		if (searchScope instanceof Resource) {
			allRefSettings = EcoreUtil.UsageCrossReferencer.find(selectedObjClass, (Resource) searchScope);
		}
		else if (searchScope instanceof ResourceSet) {
			allRefSettings = EcoreUtil.UsageCrossReferencer.find(selectedObjClass, (ResourceSet) searchScope);
		}


		/*- if refs of the EClass are found, then fill the plantUMLBuffer with the appropriate contents */

		if (allRefSettings.size() > 0) {

			for (final EStructuralFeature.Setting refSetting : allRefSettings) {

				if (refSetting.getEObject() instanceof EReference
						&& !isElementContained(result, (EReference) refSetting.getEObject())) {
					result.add((EReference) refSetting.getEObject());
				}
			}

			for (final EReference eReference : result) {

				final EClassifier eType = eReference.getEContainingClass();

				addContentToBuffer(plantumlBuffer, "class \"" + eType.getName() + "\""
						+ (eReference.isContainment() ? "" : "") + "    [[" + getUUID(eType) + "]]");

				plantumlBuffer.append(System.getProperty("line.separator"));
			}

			for (final EReference eReference : result) {

				final EClassifier referenceClasseType = eReference.getEContainingClass();


				final String name = eReference.getName();

				final boolean containment = eReference.isContainment();

				final boolean many = eReference.isMany();

				if (containment) {

					addContentToBuffer(plantumlBuffer, referenceClasseType.getName() + "*-- "
							+ selectedObjClass.getName() + ":" + (many ? " (0-*) " : "") + name + " >");
				}
				else {
					addContentToBuffer(plantumlBuffer, referenceClasseType.getName() + "--> "
							+ selectedObjClass.getName() + ":" + (many ? " (0-*) " : "") + name);

				}

				plantumlBuffer.append(System.getProperty("line.separator"));
			}
		}
		else {

			/*- no references for this EClass are found */
			addContentToBuffer(plantumlBuffer,
					"class \"" + selectedObjClass.getName() + "\"" + "    [[" + getUUID(selectedObjClass) + "]]");
			plantumlBuffer.append(System.getProperty("line.separator"));

		}
	}


	private void populateEClassContainerInfo(final StringBuffer plantumlBuffer, final EClass selectedObjClass) {
		final List<EReference> result = new ArrayList<EReference>();


		final Object searchScope = (selectedObjClass.eResource().getResourceSet() == null)
				? selectedObjClass.eResource() : selectedObjClass.eResource().getResourceSet();

		/*- getting all the elements from the selected objects EClass --> by supplying the EResource of the corresponding EClass (i.e. ecore or Xcore) */

		Collection<Setting> allRefSettings = null;
		if (searchScope instanceof Resource) {
			allRefSettings = EcoreUtil.UsageCrossReferencer.find(selectedObjClass, (Resource) searchScope);
		}
		else if (searchScope instanceof ResourceSet) {
			allRefSettings = EcoreUtil.UsageCrossReferencer.find(selectedObjClass, (ResourceSet) searchScope);
		}


		/*- if refs of the EClass are found, then fill the plantUMLBuffer with the appropriate contents */

		if (allRefSettings.size() > 0) {

			for (final EStructuralFeature.Setting refSetting : allRefSettings) {

				if (refSetting.getEObject() instanceof EReference
						&& !isElementContained(result, (EReference) refSetting.getEObject())) {
					if (((EReference) refSetting.getEObject()).isContainment()) {
						result.add((EReference) refSetting.getEObject());
					}
				}
			}

			for (final EReference eReference : result) {

				/*- only containment reference information should be displayed --> other info should be ignored */

				final EClassifier eType = eReference.getEContainingClass();

				addContentToBuffer(plantumlBuffer, "class \"" + eType.getName() + "\""
						+ (eReference.isContainment() ? "" : "") + "    [[" + getUUID(eType) + "]]");

				plantumlBuffer.append(System.getProperty("line.separator"));

			}

			for (final EReference eReference : result) {

				final EClassifier referenceClasseType = eReference.getEContainingClass();


				final String name = eReference.getName();

				final boolean containment = eReference.isContainment();

				final boolean many = eReference.isMany();

				if (containment) {

					addContentToBuffer(plantumlBuffer, referenceClasseType.getName() + "*-- "
							+ selectedObjClass.getName() + ":" + (many ? " (0-*) " : "") + name + " >");
				}
				else {
					addContentToBuffer(plantumlBuffer, referenceClasseType.getName() + "--> "
							+ selectedObjClass.getName() + ":" + (many ? " (0-*) " : "") + name);

				}

				plantumlBuffer.append(System.getProperty("line.separator"));
			}
		}
		else {

			/*- no references for this EClass are found */
			addContentToBuffer(plantumlBuffer,
					"class \"" + selectedObjClass.getName() + "\"" + "    [[" + getUUID(selectedObjClass) + "]]");
			plantumlBuffer.append(System.getProperty("line.separator"));

		}
	}


	private boolean isElementContained(final List<EReference> l, final EReference eRef) {
		boolean isElementFound = false;
		final Iterator<EReference> it = l.iterator();
		while (it.hasNext() && !isElementFound) {
			final EReference isContained = it.next();
			if (isContained.equals(eRef)) {
				isElementFound = true;
			}
		}
		return isElementFound;
	}


	private String getUUID(final EClassifier eClass) {

		if (this.id2ObjectsMap.containsValue(eClass)) {


			for (final String key : this.id2ObjectsMap.keySet()) {
				if (eClass == this.id2ObjectsMap.get(key)) {
					return key;
				}
			}
		}

		final String uuid = EcoreUtil.generateUUID();

		this.id2ObjectsMap.put(uuid, eClass);

		return uuid;

	}


}
